"""
Module implementing reference's object management logic
"""


class ObjectManager:
    """
    Manage objects listing for a given repository
    """
    # TODO: Implement


class WriteableObjectManager:
    """
    Manage writeable objects for a given repository
    """
    # TODO: Implement
